from django.conf import settings


# Ensure request context processor is active. Handle all styles of the setting.
try:
    tcp = settings.TEMPLATES[0]["OPTIONS"]["context_processors"]
except (AttributeError, KeyError):
    try:
        tcp = settings.TEMPLATE_CONTEXT_PROCESSORS
    except AttributeError:
        tcp = []

if ("django.core.context_processors.request" not in tcp) \
    and ("django.template.context_processors.request" not in tcp):
    raise RuntimeError(
        "django.template.context_processors.request is required"
    )

import ultracache.monkey



## -*- coding: utf-8 -*-

from django import template
from django.conf import settings
from django.core.cache import cache
from django.http.cookie import SimpleCookie
from django.test import TestCase
from django.test.client import Client, RequestFactory
from django.test.utils import override_settings
from django.urls import reverse

from ultracache.tests.models import DummyModel, DummyForeignModel, \
    DummyOtherModel
from ultracache.tests import views
from ultracache.tests.utils import dummy_proxy


class TemplateTagsTestCase(TestCase):
    if "django.contrib.sites" in settings.INSTALLED_APPS:
        fixtures = ["sites.json"]

    @classmethod
    def setUpClass(cls):
        super(TemplateTagsTestCase, cls).setUpClass()
        cls.factory = RequestFactory()
        cls.request = cls.factory.get('/')
        cache.clear()
        dummy_proxy.clear()

    if "django.contrib.sites" in settings.INSTALLED_APPS:
        def test_sites(self):
            from django.contrib.sites.models import Site

            first_site = Site.objects.all().first()
            second_site = Site.objects.all().last()

            # Caching on same site
            t = template.Template("{% load ultracache_tags %}\
                {% ultracache 1200 'test_ultracache' %}1{% endultracache %}"
            )
            context = template.Context({'request' : self.request})
            result1 = t.render(context)
            t = template.Template("{% load ultracache_tags %}\
                {% ultracache 1200 'test_ultracache' %}2{% endultracache %}"
            )
            context = template.Context({'request' : self.request})
            result2 = t.render(context)
            self.failUnlessEqual(result1, result2)

            # Caching on different sites
            t = template.Template("{% load ultracache_tags %}\
                {% ultracache 1200 'test_ultracache' %}1{% endultracache %}"
            )
            context = template.Context({'request' : self.request})
            result1 = t.render(context)
            with override_settings(SITE_ID=second_site.id):
                t = template.Template("{%% load ultracache_tags %%}\
                    {%% ultracache 1200 'test_ultracache' %%}%s{%% endultracache %%}" % second_site.id
                )
                context = template.Context({'request' : self.request})
                result2 = t.render(context)
                self.failIfEqual(result1, result2)

    def test_variables(self):
        # Check that undefined variables do not break caching
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_undefined' aaa %}1{% endultracache %}"
        )
        context = template.Context({'request' : self.request})
        result1 = t.render(context)
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_undefined' bbb %}2{% endultracache %}"
        )
        context = template.Context({'request' : self.request})
        result2 = t.render(context)
        self.failUnlessEqual(result1, result2)

        # Check that translation proxies are valid variables
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_xlt' _('aaa') %}1{% endultracache %}"
        )
        context = template.Context({'request' : self.request})
        result1 = t.render(context)
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_xlt' _('aaa') %}2{% endultracache %}"
        )
        context = template.Context({'request' : self.request})
        result2 = t.render(context)
        self.failUnlessEqual(result1, result2)

        # Check that large integer variables do not break caching
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_large' 565417614189797377 %}abcde{% endultracache %}"
        )
        context = template.Context({'request' : self.request})
        result1 = t.render(context)
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_large' 565417614189797377 %}abcde{% endultracache %}"
        )
        context = template.Context({'request' : self.request})
        result2 = t.render(context)
        self.failUnlessEqual(result1, result2)

    def test_context_without_request(self):
        t = template.Template("{% load ultracache_tags %}\
            {% ultracache 1200 'test_ultracache_undefined' aaa %}abcde{% endultracache %}"
        )
        context = template.Context()
        self.assertRaises(KeyError, t.render, context)

    def test_invalidation(self):
        """Directly render template
        """
        one = DummyModel.objects.create(title='One', code='one')
        two = DummyModel.objects.create(title='Two', code='two')
        three = DummyForeignModel.objects.create(title='Three', points_to=one, code='three')
        four = DummyOtherModel.objects.create(title='Four', code='four')
        # The counter is used to track the iteration that a cached block was
        # last rendered.
        t = template.Template("""{% load ultracache_tags ultracache_test_tags %}
            {% ultracache 1200 'test_ultracache_invalidate_outer' %}
                    counter outer = {{ counter }}
                {% ultracache 1200 'test_ultracache_invalidate_one' %}
                    title = {{ one.title }}
                    counter one = {{ counter }}
                {% endultracache %}
                {% ultracache 1200 'test_ultracache_invalidate_two' %}
                    title = {{ two.title }}
                    counter two = {{ counter }}
                {% endultracache %}
                {% ultracache 1200 'test_ultracache_invalidate_three' %}
                    title = {{ three.title }}
                    {{ three.points_to.title }}
                    counter three = {{ counter }}
                {% endultracache %}
                {% ultracache 1200 'test_ultracache_invalidate_render_view' %}
                    {% render_view 'render-view' %}
                {% endultracache %}
                {% ultracache 1200 'test_ultracache_invalidate_include %}
                    {% include "ultracache/include_me.html" %}
                {% endultracache %}
            {% endultracache %}"""
        )

        # Initial render
        request = self.factory.get('/aaa/')
        context = template.Context({
            'request' : request,
            'one': one,
            'two': two,
            'three': three,
            'counter': 1
        })
        result = t.render(context)
        dummy_proxy.cache(request, result)
        self.failUnless('title = One' in result)
        self.failUnless('title = Two' in result)
        self.failUnless('counter outer = 1' in result)
        self.failUnless('counter one = 1' in result)
        self.failUnless('counter two = 1' in result)
        self.failUnless('counter three = 1' in result)
        self.failUnless('render_view = One' in result)
        self.failUnless('include = One' in result)
        self.failUnless(dummy_proxy.is_cached('/aaa/'))

        # Change object one
        one.title = 'Onxe'
        one.save()
        request = self.factory.get('/bbb/')
        context = template.Context({
            'request' : request,
            'one': one,
            'two': two,
            'three': three,
            'counter': 2
        })
        result = t.render(context)
        dummy_proxy.cache(request, result)
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Two' in result)
        self.failUnless('counter outer = 2' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 1' in result)
        self.failUnless('counter three = 2' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failIf(dummy_proxy.is_cached('/aaa/'))

        # Change object two
        two.title = 'Twxo'
        two.save()
        request = self.factory.get('/ccc/')
        context = template.Context({
            'request' : request,
            'one': one,
            'two': two,
            'three': three,
            'counter': 3
        })
        result = t.render(context)
        dummy_proxy.cache(request, result)
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('counter outer = 3' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 2' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failIf(dummy_proxy.is_cached('/bbb/'))

        # Change object three
        three.title = 'Threxe'
        three.save()
        request = self.factory.get('/ddd/')
        context = template.Context({
            'request' : request,
            'one': one,
            'two': two,
            'three': three,
            'counter': 4
        })
        result = t.render(context)
        dummy_proxy.cache(request, result)
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('counter outer = 4' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failIf(dummy_proxy.is_cached('/ccc/'))

        # Add a DummyOtherModel object five
        five = DummyOtherModel.objects.create(title='Five', code='five')
        request = self.factory.get('/eee/')
        context = template.Context({
            'request' : request,
            'one': one,
            'two': two,
            'three': three,
            'counter': 5
        })
        result = t.render(context)
        dummy_proxy.cache(request, result)
        # RenderView is only view aware of DummyOtherModel. That means
        # test_ultracache_invalidate_outer and
        # test_ultracache_invalidate_render_view are expired.
        self.failUnless('render_view = Five' in result)
        self.failUnless('counter outer = 5' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failIf(dummy_proxy.is_cached('/ddd/'))

        # Delete object two
        two.delete()
        request = self.factory.get('/fff/')
        context = template.Context({
            'request' : request,
            'one': one,
            'two': None,
            'three': three,
            'counter': 6
        })
        result = t.render(context)
        dummy_proxy.cache(request, result)
        self.failUnless('title = Onxe' in result)
        self.failIf('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('counter outer = 6' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 6' in result)
        self.failUnless('counter three = 4' in result)
        self.failIf(dummy_proxy.is_cached('/eee/'))


class DecoratorTestCase(TestCase):
    if "django.contrib.sites" in settings.INSTALLED_APPS:
        fixtures = ["sites.json"]

    def setUp(self):
        super(DecoratorTestCase, self).setUp()
        cache.clear()
        dummy_proxy.clear()

    def test_method(self):
        """Render template through a view with get method decorated with
        cached_get."""
        one = DummyModel.objects.create(title='One', code='one')
        two = DummyModel.objects.create(title='Two', code='two')
        three = DummyForeignModel.objects.create(title='Three', points_to=one, code='three')
        four = DummyModel.objects.create(title='Four', code='four')
        five = DummyModel.objects.create(title='Five', code='five')
        url = reverse('method-cached-view')

        # Initial render
        cache.set("counter", 1)
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.assertEqual(response.status_code, 200)
        self.failUnless('title = One' in result)
        self.failUnless('title = Two' in result)
        self.failUnless('title = Three' in result)
        self.failUnless('render_view = One' in result)
        self.failUnless('include = One' in result)
        self.failUnless('counter one = 1' in result)
        self.failUnless('counter two = 1' in result)
        self.failUnless('counter three = 1' in result)
        self.failUnless('counter four = 1' in result)
        self.failUnless('title = Four' in result)

        # Change object one
        cache.set("counter", 2)
        one.title = 'Onxe'
        one.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Two' in result)
        self.failUnless('title = Three' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 1' in result)
        self.failUnless('counter three = 2' in result)
        self.failUnless('counter four = 2' in result)
        self.failUnless('title = Four' in result)

        # Change object two
        cache.set("counter", 3)
        two.title = 'Twxo'
        two.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Three' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 2' in result)
        self.failUnless('counter four = 3' in result)
        self.failUnless('title = Four' in result)

        # Change object three
        cache.set("counter", 4)
        three.title = 'Threxe'
        three.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('counter four = 4' in result)
        self.failUnless('title = Four' in result)

        # Change object four
        cache.set("counter", 5)
        four.title = 'Fouxr'
        four.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('counter four = 5' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('title = Fouxr' in result)
        self.failIf('title = Four' in result)

        # Change object five. This object is never accessed in the template,
        # only get_context_data of CachedView. "counter four" and "counter
        # five" are under cached_get and not in any ultracache tag, and are
        # thus the only counters incremented.
        cache.set("counter", 6)
        five.title = 'Fivxe'
        five.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('counter four = 6' in result)
        self.failUnless('counter five = 6' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('title = Fouxr' in result)
        self.failIf('title = Four' in result)

    def test_class(self):
        """Render template through a view decorated with ultracache
        """
        one = DummyModel.objects.create(title='One', code='one')
        two = DummyModel.objects.create(title='Two', code='two')
        three = DummyForeignModel.objects.create(title='Three', points_to=one, code='three')
        four = DummyModel.objects.create(title='Four', code='four')
        five = DummyModel.objects.create(title='Five', code='five')
        url = reverse('class-cached-view')

        # Initial render
        cache.set("counter", 1)
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.assertEqual(response.status_code, 200)
        self.failUnless('title = One' in result)
        self.failUnless('title = Two' in result)
        self.failUnless('title = Three' in result)
        self.failUnless('render_view = One' in result)
        self.failUnless('include = One' in result)
        self.failUnless('counter one = 1' in result)
        self.failUnless('counter two = 1' in result)
        self.failUnless('counter three = 1' in result)
        self.failUnless('counter four = 1' in result)
        self.failUnless('title = Four' in result)

        # Change object one
        cache.set("counter", 2)
        one.title = 'Onxe'
        one.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Two' in result)
        self.failUnless('title = Three' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 1' in result)
        self.failUnless('counter three = 2' in result)
        self.failUnless('counter four = 2' in result)
        self.failUnless('title = Four' in result)

        # Change object two
        cache.set("counter", 3)
        two.title = 'Twxo'
        two.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Three' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 2' in result)
        self.failUnless('counter four = 3' in result)
        self.failUnless('title = Four' in result)

        # Change object three
        cache.set("counter", 4)
        three.title = 'Threxe'
        three.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('counter four = 4' in result)
        self.failUnless('title = Four' in result)

        # Change object four
        cache.set("counter", 5)
        four.title = 'Fouxr'
        four.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('counter four = 5' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('title = Fouxr' in result)
        self.failIf('title = Four' in result)

        # Change object five. This object is never accessed in the template,
        # only get_context_data of CachedView. "counter four" and "counter
        # five" are under cached_get and not in any ultracache tag, and are
        # thus the only counters incremented.
        cache.set("counter", 6)
        five.title = 'Fivxe'
        five.save()
        response = self.client.get(url)
        result = response.content.decode("utf-8")
        self.failUnless('title = Onxe' in result)
        self.failIf('title = One' in result)
        self.failUnless('title = Twxo' in result)
        self.failIf('title = Two' in result)
        self.failUnless('title = Threxe' in result)
        self.failIf('title = Three' in result)
        self.failUnless('counter one = 2' in result)
        self.failUnless('counter two = 3' in result)
        self.failUnless('counter three = 4' in result)
        self.failUnless('counter four = 6' in result)
        self.failUnless('counter five = 6' in result)
        self.failUnless('render_view = Onxe' in result)
        self.failUnless('include = Onxe' in result)
        self.failUnless('title = Fouxr' in result)
        self.failIf('title = Four' in result)

    def test_header(self):
        """Test that decorator preserves headers
        """
        url = reverse('cached-header-view')

        # Initial render
        response = self.client.get(url)
        # Django 4 deprecates _headers and introduces headers
        if hasattr(response, "headers"):
            self.assertEqual(response.headers['content-type'], 'application/json')
            self.assertEqual(response.headers['foo'], 'bar')
        else:
            self.assertEqual(response._headers['content-type'], ('Content-Type', 'application/json'))
            self.assertEqual(response._headers['foo'], ('foo', 'bar'))

        # Second pass is cached
        response = self.client.get(url)
        # Django 4 deprecates _headers and introduces headers
        if hasattr(response, "headers"):
            self.assertEqual(response.headers['content-type'], 'application/json')
            self.assertEqual(response.headers['foo'], 'bar')
        else:
            self.assertEqual(response._headers['content-type'], ('Content-Type', 'application/json'))
            self.assertEqual(response._headers['foo'], ('foo', 'bar'))

    def test_cache_busting(self):
        """Test cache busting with and without random querystring param
        """
        url = reverse('bustable-cached-view')
        response = self.client.get(url + '?aaa=1')
        self.failUnless('aaa=1' in response.content.decode("utf-8"))
        response = self.client.get(url + '?aaa=2')
        self.failUnless('aaa=2' in response.content.decode("utf-8"))

        url = reverse('non-bustable-cached-view')
        response = self.client.get(url + '?aaa=1')
        self.failUnless('aaa=1' in response.content.decode("utf-8"))
        response = self.client.get(url + '?aaa=2')
        self.failIf('aaa=2' in response.content.decode("utf-8"))



"""Template tag used in unit tests"""

import re

from django import template
from django.conf import settings
from django.http import HttpResponse
from django.template.response import TemplateResponse
from django.urls import reverse, resolve, get_script_prefix

register = template.Library()


@register.tag
def render_view(parser, token):
    """{% render_view view_name %}"""
    tokens = token.split_contents()
    if len(tokens) != 2:
        raise template.TemplateSyntaxError(
            "render_view view_name %}"
        )
    return RenderViewNode(tokens[1])


class RenderViewNode(template.Node):

    def __init__(self, view_name):
        self.view_name = template.Variable(view_name)

    def render(self, context):
        view_name = self.view_name.resolve(context)
        url = reverse(view_name)
        # Resolve needs any possible prefix removed
        url = re.sub(r"^%s" % get_script_prefix().rstrip("/"), "", url)
        view, args, kwargs = resolve(url)
        # Call the view. Let any error propagate.
        request = context["request"]
        result = view(request, *args, **kwargs)
        if isinstance(result, TemplateResponse):
            # The result of a generic view
            result.render()
            html = result.rendered_content
        elif isinstance(result, HttpResponse):
            # Old-school view
            html = result.content
        return html



import json
try:
    from urllib.parse import urlparse
    from urllib.parse import quote
except ImportError:
    from urlparse import urlparse
    from urllib import quote

from celery import shared_task
try:
    import pika
    DO_TASK = True
except ImportError:
    DO_TASK = False

from django.conf import settings


@shared_task(max_retries=3, ignore_result=True)
def broadcast_purge(path, headers=None):
    if not DO_TASK:
        raise RuntimeError("Library pika>=0.11,<1.0 not found")

    try:
        url = settings.ULTRACACHE["rabbitmq-url"]
    except (AttributeError, KeyError):
        # Use same host as celery. Pika requires the path to be url
        # encoded. A typical broker URL setting remains effectively
        # unchanged but as soon as sub-paths are encountered this encoding
        # becomes necessary.
        parsed = urlparse(settings.CELERY_BROKER_URL)
        url = "%s://%s/%s" % (
            parsed.scheme,
            parsed.netloc,
            quote(parsed.path[1:], safe="")
        )
    connection = pika.BlockingConnection(pika.URLParameters(url))
    channel = connection.channel()
    channel.exchange_declare(exchange="purgatory", exchange_type="fanout")
    channel.basic_publish(
        exchange="purgatory",
        routing_key="",
        body=json.dumps({"path": path, "headers": headers or {}})
    )
    connection.close()
    return True



import django

from rest_framework import viewsets, serializers

from ultracache.tests.models import DummyModel


class DummySerializer(serializers.ModelSerializer):

    class Meta:
        model = DummyModel
        if not django.get_version().startswith("1.6"):
            fields = "__all__"


class DummyViewSet(viewsets.ModelViewSet):
    queryset = DummyModel.objects.all()
    serializer_class = DummySerializer



from ultracache import _thread_locals


def _cleanup():
    if hasattr(_thread_locals, "ultracache_recorder"):
        delattr(_thread_locals, "ultracache_recorder")
    if hasattr(_thread_locals, "ultracache_attr_marker"):
        delattr(_thread_locals, "ultracache_attr_marker")


class UltraCacheMiddleware(object):
    """Middleware to ensure thread locals is cleaned up.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.process_request(request)
        try:
            response = self.get_response(request)
        except Exception as e:
            self.process_exception(request, e)
            raise
        return self.process_response(request, response)

    def process_request(self, request):
        setattr(_thread_locals, "ultracache_recorder", [])

    def process_response(self, request, response):
        _cleanup()
        return response

    def process_exception(self, request, exception):
        _cleanup()



"""Monkey patch template variable resolution so we can recognize which objects
are covered within a containing caching template tag. The patch is based on
Django 1.11 but is backwards compatible with 1.9."""

import hashlib
import inspect
import pickle
import types
from collections import OrderedDict

from django.core.cache import cache
from django.db.models import Model, Manager
from django.template.base import Variable, VariableDoesNotExist
from django.template.context import BaseContext
from django.contrib.contenttypes.models import ContentType
from django.conf import settings

from ultracache import _thread_locals
from ultracache.utils import cache_meta, get_current_site_pk

try:
    from django.template.base import logger
except ImportError:
    logger = None


def my_resolve_lookup(self, context):
        """
        Performs resolution of a real variable (i.e. not a literal) against the
        given context.

        As indicated by the method's name, this method is an implementation
        detail and shouldn"t be called by external code. Use Variable.resolve()
        instead.
        """
        current = context
        try:  # catch-all for silent variable failures
            for bit in self.lookups:
                try:  # dictionary lookup
                    current = current[bit]
                    # ValueError/IndexError are for numpy.array lookup on
                    # numpy < 1.9 and 1.9+ respectively
                except (TypeError, AttributeError, KeyError, ValueError, IndexError):
                    try:  # attribute lookup
                        # Don"t return class attributes if the class is the context:
                        if isinstance(current, BaseContext) and getattr(type(current), bit):
                            raise AttributeError
                        current = getattr(current, bit)
                    except (TypeError, AttributeError) as e:
                        # Reraise an AttributeError raised by a @property
                        if (isinstance(e, AttributeError) and
                                not isinstance(current, BaseContext) and bit in dir(current)):
                            raise
                        try:  # list-index lookup
                            current = current[int(bit)]
                        except (IndexError,  # list index out of range
                                ValueError,  # invalid literal for int()
                                KeyError,    # current is a dict without `int(bit)` key
                                TypeError):  # unsubscriptable object
                            raise VariableDoesNotExist("Failed lookup for key "
                                                       "[%s] in %r",
                                                       (bit, current))  # missing attribute
                if callable(current):
                    if getattr(current, "do_not_call_in_templates", False):
                        pass
                    elif getattr(current, "alters_data", False):
                        try:
                            current = context.template.engine.string_if_invalid
                        except AttributeError:
                            current = settings.TEMPLATE_STRING_IF_INVALID
                    else:
                        try:  # method call (assuming no args required)
                            current = current()
                        except TypeError:
                            try:
                                inspect.getcallargs(current)
                            except TypeError:  # arguments *were* required
                                current = context.template.engine.string_if_invalid  # invalid method call
                            else:
                                raise
                elif isinstance(current, Model):
                    if ("request" in context) and hasattr(context["request"], "_ultracache"):
                        # get_for_model itself is cached
                        ct = ContentType.objects.get_for_model(current.__class__)
                        context["request"]._ultracache.append((ct.id, current.pk))

        except Exception as e:
            template_name = getattr(context, "template_name", None) or "unknown"
            if logger is not None:
                logger.debug(
                    "Exception while resolving variable \"%s\" in template \"%s\".",
                    bit,
                    template_name,
                    exc_info=True,
                )

            if getattr(e, "silent_variable_failure", False):
                current = context.template.engine.string_if_invalid
            else:
                raise

        return current

Variable._resolve_lookup = my_resolve_lookup


"""If Django Rest Framework is installed patch a few mixins. Serializers are
conceptually the same as templates but make it even easier to track objects."""
try:
    from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
    from rest_framework.response import Response
    from rest_framework.serializers import Serializer, ListSerializer
    HAS_DRF = True
except ImportError:
    HAS_DRF = False


def drf_cache(func):

    def wrapped(context, request, *args, **kwargs):
        viewsets = settings.ULTRACACHE.get("drf", {}).get("viewsets", {})
        dotted_name =  context.__module__ + "." + context.__class__.__name__
        do_cache = (dotted_name in viewsets) or (context.__class__ in viewsets) or ("*" in viewsets)

        if do_cache:
            li = [request.get_full_path()]
            viewset_settings = viewsets.get(dotted_name, {}) \
                or viewsets.get(context.__class__, {}) \
                or viewsets.get("*", {})
            evaluate = viewset_settings.get("evaluate", None)
            if evaluate is not None:
                if callable(evaluate):
                    li.append(evaluate(context, request))
                else:
                    li.append(eval(evaluate))

            if "django.contrib.sites" in settings.INSTALLED_APPS:
                li.append(get_current_site_pk(request))

            s = ":".join([str(l) for l in li])
            cache_key = hashlib.md5(s.encode("utf-8")).hexdigest()

            cached = cache.get(cache_key, None)
            if cached is not None:
                response = Response(pickle.loads(cached["content"]))

                # Headers has a non-obvious format
                for k, v in cached["headers"].items():
                    response[v[0]] = v[1]

                return response

        if not hasattr(request, "_ultracache"):
            setattr(request, "_ultracache", [])
            setattr(request, "_ultracache_cache_key_range", [])

        response = func(context, request, *args, **kwargs)

        if do_cache:
            cache_meta(_thread_locals.ultracache_recorder, cache_key, request=request)
            response = context.finalize_response(request, response, *args, **kwargs)
            response.render()
            timeout = viewset_settings.get("timeout", 300)
            headers = getattr(response, "_headers", {})
            cache.set(
                cache_key,
                {"content": pickle.dumps(response.data), "headers": headers},
                timeout
            )
            return response

        else:
            return response

    return wrapped


def _serializer(func):
    # Helper decorator for Serializer

    def wrapped(context, instance):
        request = context.context["request"]
        if hasattr(request, "_ultracache") and isinstance(instance, Model):
            ct = ContentType.objects.get_for_model(instance.__class__)
            request._ultracache.append((ct.id, instance.pk))
        return func(context, instance)

    return wrapped


def _listserializer(func):
    # Helper decorator for ListSerializer

    def wrapped(context, data):
        request = context.context["request"]
        if hasattr(request, "_ultracache"):
            iterable = data.all() if isinstance(data, Manager) else data
            for obj in iterable:
                if isinstance(obj, Model):
                    ct = ContentType.objects.get_for_model(obj.__class__)
                    request._ultracache.append((ct.id, obj.pk))
        return func(context, data)

    return wrapped


if HAS_DRF:
    ListModelMixin.list = drf_cache(ListModelMixin.list)
    RetrieveModelMixin.retrieve = drf_cache(RetrieveModelMixin.retrieve)
    Serializer.to_representation = _serializer(Serializer.to_representation)
    ListSerializer.to_representation = _listserializer(ListSerializer.to_representation)


# Dark magic enables us to record object access. This enables us keep track of
# objects accessed in any Python code and not just in the templates being
# rendered.

def my__getattribute__(self, name):
    if hasattr(_thread_locals, "ultracache_recorder") and \
        not hasattr(_thread_locals, "_ultracache_attr_marker"):
        setattr(_thread_locals, "_ultracache_attr_marker", 1)
        if hasattr(self, "pk"):
            # get_for_model itself is cached
            ct = ContentType.objects.get_for_model(self.__class__)
            _thread_locals.ultracache_recorder.append((ct.id, self.pk))
        delattr(_thread_locals, "_ultracache_attr_marker")
    return super(Model, self).__getattribute__(name)

Model.__getattribute__ = my__getattribute__



from django.core.cache import cache
from django.views.generic.base import TemplateView

from ultracache.decorators import cached_get, ultracache
from ultracache.tests.models import DummyModel, DummyForeignModel, \
    DummyOtherModel


class RenderView(TemplateView):
    """Simple view that renders a dummy model.
    """
    template_name = "ultracache/render_view.html"

    def get_context_data(self, **kwargs):
        context = super(RenderView, self).get_context_data(**kwargs)
        context["one"] = DummyModel.objects.get(code="one")
        try:
            context["four"] = DummyOtherModel.objects.get(code="four")
        except DummyOtherModel.DoesNotExist:
            pass
        try:
            context["five"] = DummyOtherModel.objects.get(code="five")
        except DummyOtherModel.DoesNotExist:
            pass
        return context


class BaseCachedView(TemplateView):
    template_name = "ultracache/cached_view.html"

    def get_context_data(self, **kwargs):
        context = super(BaseCachedView, self).get_context_data(**kwargs)
        context["one"] = DummyModel.objects.get(code="one")
        context["two"] = DummyModel.objects.get(code="two")
        context["three"] = DummyForeignModel.objects.get(code="three")
        context["four"] = DummyModel.objects.get(code="four")

        # The counter is used to track the iteration that a cached block was
        # last rendered. Setting the value on the cache is the easiest way to
        # influence the counter from a test.
        context["counter"] = cache.get("counter", 1)

        # We never actually do anything with five, but cached_get is aware of it
        dc = DummyModel.objects.get(code="five")

        return context


class MethodCachedView(BaseCachedView):

    @cached_get(300, "request.is_secure()", 456)
    def get(self, *args, **kwargs):
        return super(MethodCachedView, self).get(*args, **kwargs)


@ultracache(300, "request.is_secure()", 456)
class ClassCachedView(BaseCachedView):
    pass


class CachedHeaderView(TemplateView):
    template_name = "ultracache/cached_header_view.html"
    content_type = "application/json"

    @cached_get(300)
    def get(self, *args, **kwargs):
        response = super(CachedHeaderView, self).get(*args, **kwargs)
        response["foo"] = "bar"
        return response


class BustableCachedView(TemplateView):
    template_name = "ultracache/bustable_cached_view.html"

    @cached_get(300)
    def get(self, *args, **kwargs):
        return super(BustableCachedView, self).get(*args, **kwargs)


class NonBustableCachedView(TemplateView):
    template_name = "ultracache/non_bustable_cached_view.html"

    @cached_get(300, "request.path_info")
    def get(self, *args, **kwargs):
        return super(NonBustableCachedView, self).get(*args, **kwargs)



from collections import OrderedDict


class DummyProxy(dict):

    def make_key(self, path, headers=None):
        key = path
        if headers is not None:
            key += str(frozenset(sorted(headers.items())))
        return key

    def cache(self, request, value):
        headers = {k[5:].replace("_", "-").lower(): v for \
            k, v in request.META.items() if k.startswith("HTTP_")}
        key = self.make_key(request.get_full_path(), headers)
        self[key] = value

    def is_cached(self, path, headers=None):
        # The test framework sends an empty cookie with each request. Avoid
        # copy pasta in the individual tests and just add that header here.
        if headers is None:
            headers = {u"cookie": u""}
        key = self.make_key(path, headers)
        return key in self

    def purge(self, path, headers=None):
        key = self.make_key(path, headers)
        if key in self:
            del self[key]

dummy_proxy = DummyProxy()


def dummy_purger(path, headers=None):
    dummy_proxy.purge(path, headers=headers)



import requests

from django.conf import settings


def broadcast(path, headers=None):
    # The preferred methody requires RabbitMQ and celery being installed and
    # configured.
    from ultracache.tasks import broadcast_purge
    broadcast_purge.delay(path, headers)


def varnish(path, headers=None):
    # See https://www.varnish-software.com/static/book/Cache_invalidation.html
    loc = settings.ULTRACACHE["purge"]["method"]["url"].rstrip("/") + "/" \
        + path.lstrip("/")
    try:
        r = requests.request("PURGE", loc, timeout=1, headers=headers or {})
    except requests.exceptions.RequestException:
        pass


def nginx(path, headers=None):
    # See https://github.com/FRiCKLE/ngx_cache_purge

    # Simplest case - one node
    loc = settings.ULTRACACHE["purge"]["method"]["url"].rstrip("/") + "/" \
        + path.lstrip("/")
    try:
        r = requests.request("PURGE", loc, timeout=1, headers=headers or {})
    except requests.exceptions.RequestException:
        pass



try:
    from django.urls import include, re_path
except ImportError:
    from django.conf.urls import include, re_path

from rest_framework.routers import DefaultRouter

from ultracache.tests import views, viewsets


router = DefaultRouter()
router.register(r"dummies", viewsets.DummyViewSet)

urlpatterns = [
    re_path(r"^api/", include(router.urls)),
    re_path(
        r"^render-view/$",
        views.RenderView.as_view(),
        name="render-view"
    ),
    re_path(
        r"^method-cached-view/$",
        views.MethodCachedView.as_view(),
        name="method-cached-view"
    ),
    re_path(
        r"^class-cached-view/$",
        views.ClassCachedView.as_view(),
        name="class-cached-view"
    ),
    re_path(
        r"^cached-header-view/$",
        views.CachedHeaderView.as_view(),
        name="cached-header-view"
    ),
    re_path(
        r"^bustable-cached-view/$",
        views.BustableCachedView.as_view(),
        name="bustable-cached-view"
    ),
    re_path(
        r"^non-bustable-cached-view/$",
        views.NonBustableCachedView.as_view(),
        name="non-bustable-cached-view"
    ),
]



from django.conf import settings
from django.core.cache import cache
from django.test import TestCase

from ultracache.utils import Ultracache
from ultracache.tests.models import DummyModel


class UtilsTestCase(TestCase):
    if "django.contrib.sites" in settings.INSTALLED_APPS:
        fixtures = ["sites.json"]

    def setUp(self):
        super(UtilsTestCase, self).setUp()
        cache.clear()

    def test_context_manager_like_thing(self):
        one = DummyModel.objects.create(title="One", code="one")
        two = DummyModel.objects.create(title="Two", code="two")

        # Caching with object one
        uc = Ultracache(3600, "a", "b")
        self.failIf(uc)
        uc.cache(one.title)

        uc = Ultracache(3600, "a", "b")
        self.failUnless(uc)
        self.assertEqual(uc.cached, one.title)

        one.title = "Onex"
        one.save()

        uc = Ultracache(3600, "a", "b")
        self.failIf(uc)

        # Caching with object two. Ensure object one doesn't bleed into this
        # section.
        uc = Ultracache(3600, "c", "d")
        self.failIf(uc)
        uc.cache(two.title)

        uc = Ultracache(3600, "c", "d")
        self.failUnless(uc)
        self.assertEqual(uc.cached, two.title)

        two.title = "Onez"
        one.save()
        uc = Ultracache(3600, "c", "d")
        self.failUnless(uc)

        two.title = "Twox"
        two.save()

        uc = Ultracache(3600, "c", "d")
        self.failIf(uc)



import os


DEBUG = True
TEMPLATE_DEBUG = DEBUG

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
    }
}

ROOT_URLCONF = "ultracache.tests.urls"

INSTALLED_APPS = (
    "test_without_migrations",
    "ultracache",
    # Include tests because it declares models
    "ultracache.tests",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "rest_framework"
)

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    }
}

SECRET_KEY = "SECRET_KEY"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": False,
        "OPTIONS": {
            "context_processors": ["django.template.context_processors.request"],
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                "django.template.loaders.app_directories.Loader",
            ]
        },
    },
]

MIDDLEWARE = [
    "ultracache.middleware.UltraCacheMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ULTRACACHE = {
    "purge": {"method": "ultracache.tests.utils.dummy_purger"},
    "drf": {"viewsets": {"*": {}}},
    "consider-headers": ["cookie"],
}



from setuptools import setup, find_packages

setup(
    name="django-ultracache",
    description="Drop-in replacement for Django's template fragment caching. Provides automatic cache invalidation.",
    long_description = open("README.rst", "r").read() + open("AUTHORS.rst", "r").read() + open("CHANGELOG.rst", "r").read(),
    version="2.2",
    author="Hedley Roos",
    author_email="hedleyroos@gmail.com",
    license="BSD",
    url="http://github.com/hedleyroos/django-ultracache",
    packages = find_packages(),
    dependency_links = [
    ],
    install_requires = [
        "Django>=2.0",
        "requests",
        "pika>=0.11,<1.0",
        "PyYAML>=5.0,<6.0"
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    zip_safe=False,
)



from django.db import models


class DummyModel(models.Model):
    title = models.CharField(max_length=32)
    code = models.CharField(max_length=32)


class DummyForeignModel(models.Model):
    title = models.CharField(max_length=32)
    points_to = models.ForeignKey(DummyModel, on_delete=models.CASCADE)
    code = models.CharField(max_length=32)


class DummyOtherModel(models.Model):
    title = models.CharField(max_length=32)
    code = models.CharField(max_length=32)



import threading

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.db.migrations.recorder import MigrationRecorder
from django.db.models import Model
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver

try:
    from django.utils.module_loading import import_string as importer
except ImportError:
    from django.utils.module_loading import import_by_path as importer

try:
    purger = importer(settings.ULTRACACHE["purge"]["method"])
except (AttributeError, KeyError):
    purger = None

try:
    invalidate = settings.ULTRACACHE["invalidate"]
except (AttributeError, KeyError):
    invalidate = True


@receiver(post_save)
def on_post_save(sender, **kwargs):
    """Expire ultracache cache keys affected by this object
    """
    if not invalidate:
        return
    if kwargs.get("raw", False):
        return
    if sender is MigrationRecorder.Migration:
        return
    if issubclass(sender, Model):
        obj = kwargs["instance"]
        if isinstance(obj, Model):
            # get_for_model itself is cached
            try:
                ct = ContentType.objects.get_for_model(sender)
            except RuntimeError:
                # This happens when ultracache is being used by another product
                # during a test run.
                return

            if kwargs.get("created", False):
                # Expire cache keys that contain objects of this content type
                key = "ucache-ct-%s" % ct.id
                to_delete = cache.get(key, [])
                if to_delete:
                    try:
                        cache.delete_many(to_delete)
                    except NotImplementedError:
                        for k in to_delete:
                            cache.delete(k)
                cache.delete(key)

                # Purge paths in reverse caching proxy that contain objects of
                # this content type.
                key = "ucache-ct-pth-%s" % ct.id
                if purger is not None:
                    # The key *must* be deleted first in case the purger fails
                    items = cache.get(key, [])
                    cache.delete(key)
                    for li in items:
                        purger(li[0], li[1])
                else:
                    cache.delete(key)

            else:
                # Expire cache keys
                key = "ucache-%s-%s" % (ct.id, obj.pk)
                to_delete = cache.get(key, [])
                if to_delete:
                    try:
                        cache.delete_many(to_delete)
                    except NotImplementedError:
                        for k in to_delete:
                            cache.delete(k)
                cache.delete(key)

                # Purge paths in reverse caching proxy
                key = "ucache-pth-%s-%s" % (ct.id, obj.pk)
                if purger is not None:
                    # The key *must* be deleted first in case the purger fails
                    items = cache.get(key, [])
                    cache.delete(key)
                    for li in items:
                        purger(li[0], li[1])
                else:
                    cache.delete(key)


@receiver(post_delete)
def on_post_delete(sender, **kwargs):
    """Expire ultracache cache keys affected by this object
    """
    if not invalidate:
        return
    if kwargs.get("raw", False):
        return
    if sender is MigrationRecorder.Migration:
        return
    if issubclass(sender, Model):
        obj = kwargs["instance"]
        if isinstance(obj, Model):
            # get_for_model itself is cached
            try:
                ct = ContentType.objects.get_for_model(sender)
            except RuntimeError:
                # This happens when ultracache is being used by another product
                # during a test run.
                return

            # Expire cache keys
            key = "ucache-%s-%s" % (ct.id, obj.pk)
            to_delete = cache.get(key, [])
            if to_delete:
                try:
                    cache.delete_many(to_delete)
                except NotImplementedError:
                    for k in to_delete:
                        cache.delete(k)
            cache.delete(key)

            # Invalidate paths in reverse caching proxy
            key = "ucache-pth-%s-%s" % (ct.id, obj.pk)
            if purger is not None:
                # The key *must* be deleted first in case the purger fails
                items = cache.get(key, [])
                for li in items:
                    purger(li[0], li[1])
            else:
                cache.delete(key)



from django import template
try:
    from django.utils.encoding import force_str
except ImportError:
    from django.utils.encoding import force_text as force_str
try:
    from django.utils.translation import ugettext as _
except ImportError:
    from django.utils.translation import gettext as _
from django.utils.functional import Promise
from django.templatetags.cache import CacheNode
from django.template.base import VariableDoesNotExist
from django.core.cache import cache
from django.core.cache.utils import make_template_fragment_key
from django.conf import settings

from ultracache import _thread_locals
from ultracache.utils import cache_meta, get_current_site_pk


register = template.Library()


class UltraCacheNode(CacheNode):
    """Based on Django's default cache template tag. Add SITE_ID as implicit
    vary on parameter is sites product is installed. Allow unresolvable
    variables. Allow translated strings."""

    def __init__(self, *args):
        # Django 1.7 introduced cache_name. Using different caches makes
        # invalidation difficult. It will be supported in a future version.
        try:
            super(UltraCacheNode, self).__init__(*args, cache_name=None)
        except TypeError:
            super(UltraCacheNode, self).__init__(*args)

    def render(self, context):
        try:
            expire_time = self.expire_time_var.resolve(context)
        except VariableDoesNotExist:
            raise TemplateSyntaxError(
                "ultracache tag got an unknown variable: %r" % self.expire_time_var.var
            )
        try:
            expire_time = int(expire_time)
        except (ValueError, TypeError):
            raise TemplateSyntaxError(
                "ultracache tag got a non-integer timeout value: %r" % expire_time
            )

        request = context["request"]

        # If request not GET or HEAD never cache
        if request.method.lower() not in ("get", "head"):
            return self.nodelist.render(context)

        # Set a list on the request. Django's template rendering is recursive
        # and single threaded so we can use a list to keep track of contained
        # objects.
        if not hasattr(_thread_locals, "ultracache_recorder"):
            setattr(_thread_locals, "ultracache_recorder", [])
            start_index = 0
        else:
            start_index = len(_thread_locals.ultracache_recorder)

        vary_on = []
        if "django.contrib.sites" in settings.INSTALLED_APPS:
            vary_on.append(str(get_current_site_pk(request)))

        for var in self.vary_on:
            try:
                r = var.resolve(context)
            except VariableDoesNotExist:
                pass
            if isinstance(r, Promise):
                r = force_str(r)
            vary_on.append(r)

        cache_key = make_template_fragment_key(self.fragment_name, vary_on)
        value = cache.get(cache_key)
        if value is None:
            value = self.nodelist.render(context)
            cache.set(cache_key, value, expire_time)
            cache_meta(_thread_locals.ultracache_recorder, cache_key, start_index, request=request)
        else:
            # A cached result was found. Set tuples in _ultracache manually so
            # outer template tags are aware of contained objects.
            for tu in cache.get(cache_key + "-objs", []):
                _thread_locals.ultracache_recorder.append(tu)

        return value


@register.tag("ultracache")
def do_ultracache(parser, token):
    """Based on Django's default cache template tag"""
    nodelist = parser.parse(("endultracache",))
    parser.delete_first_token()
    tokens = token.split_contents()
    if len(tokens) < 3:
        raise TemplateSyntaxError(""%r" tag requires at least 2 arguments." % tokens[0])
    return UltraCacheNode(nodelist,
        parser.compile_filter(tokens[1]),
        tokens[2], # fragment_name can"t be a variable.
        [parser.compile_filter(token) for token in tokens[3:]])



import hashlib
import types
from functools import wraps, WRAPPER_ASSIGNMENTS, partial

from django.conf import settings
from django.core.cache import cache
from django.http import HttpResponse
from django.template.response import TemplateResponse
from django.views.generic.base import TemplateResponseMixin

from ultracache import _thread_locals
from ultracache.utils import cache_meta, get_current_site_pk


def cached_get(timeout, *params):
    """Decorator applied specifically to a view's get method"""

    def decorator(view_func):
        @wraps(view_func, assigned=WRAPPER_ASSIGNMENTS)
        def _wrapped_view(view_or_request, *args, **kwargs):

            # The type of the request gets muddled when using a function based
            # decorator. We must use a function based decorator so it can be
            # used in urls.py.
            request = getattr(view_or_request, "request", view_or_request)

            # If request not GET or HEAD never cache
            if request.method.lower() not in ("get", "head"):
                return view_func(view_or_request, *args, **kwargs)

            # If request contains messages never cache
            l = 0
            try:
                l = len(request._messages)
            except (AttributeError, TypeError):
                pass
            if l:
                return view_func(view_or_request, *args, **kwargs)

            # Compute a cache key
            if isinstance(view_func, partial):
                func_name = view_func.func.__name__
            else:
                func_name = view_func.__name__
            li = [str(view_or_request.__class__), func_name]

            # request.get_full_path is implicitly added if no other request
            # path is provided. get_full_path includes the querystring and is
            # the more conservative approach but makes it trivially easy for a
            # request to bust through the cache.
            if not set(params).intersection(set((
                "request.get_full_path()", "request.path", "request.path_info"
            ))):
                li.append(request.get_full_path())

            if "django.contrib.sites" in settings.INSTALLED_APPS:
                li.append(get_current_site_pk(request))

            # Pre-sort kwargs
            keys = list(kwargs.keys())
            keys.sort()
            for key in keys:
                li.append("%s,%s" % (key, kwargs[key]))

            # Extend cache key with custom variables
            for param in params:
                if not isinstance(param, str):
                    param = str(param)
                li.append(eval(param))

            s = ":".join([str(l) for l in li])
            hashed = hashlib.md5(s.encode("utf-8")).hexdigest()
            cache_key = "ucache-%s" % hashed
            cached = cache.get(cache_key, None)
            if cached is None:
                # The get view as outermost caller may bluntly set recorder to empty
                _thread_locals.ultracache_recorder = []
                response = view_func(view_or_request, *args, **kwargs)
                content = None
                if isinstance(response, TemplateResponse):
                    content = response.render().rendered_content
                elif isinstance(response, HttpResponse):
                    content = response.content
                if content is not None:
                    # Django 4 deprecates _headers and introduces headers. Drop to private API for compatibility.
                    if hasattr(response, "headers"):
                        headers = response.headers._store
                    else:
                        headers = getattr(response, "_headers", {})
                    cache.set(
                        cache_key,
                        {"content": content, "headers": headers},
                        timeout
                    )
                    cache_meta(_thread_locals.ultracache_recorder, cache_key, request=request)
            else:
                response = HttpResponse(cached["content"])
                # Headers has a non-obvious format
                for k, v in cached["headers"].items():
                    response[v[0]] = v[1]

            return response

        return _wrapped_view
    return decorator


def ultracache(timeout, *params):
    """Decorator applied to a view class. The get method is decorated
    implicitly."""

    def decorator(cls):
        class WrappedClass(cls):
            def __init__(self, *args, **kwargs):
                super(WrappedClass, self).__init__(*args, **kwargs)

            @cached_get(timeout, *params)
            def get(self, *args, **kwargs):
                return super(WrappedClass, self).get(*args, **kwargs)

        return WrappedClass
    return decorator



import copy
import json

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import reverse

from rest_framework.test import APIRequestFactory, APIClient

from ultracache.tests.models import DummyModel
from ultracache.tests.viewsets import DummyViewSet


class DRFTestCase(TestCase):
    if "django.contrib.sites" in settings.INSTALLED_APPS:
        fixtures = ["sites.json"]

    @classmethod
    def setUpTestData(cls):
        super(DRFTestCase, cls).setUpTestData()

        cls.user_model = get_user_model()

        # Superuser
        cls.superuser = cls.user_model.objects.create(
            username="superuser",
            email="superuser@test.com",
            is_superuser=True,
            is_staff=True
        )
        cls.superuser.set_password("password")
        cls.superuser.save()

        # Staff
        cls.staff = cls.user_model.objects.create(
            username="staff",
            email="staff@test.com",
            is_staff=True
        )
        cls.staff.set_password("password")
        cls.staff.save()

        # Plain user
        cls.user = cls.user_model.objects.create(
            username="user",
            email="user@test.com"
        )
        cls.user.set_password("password")
        cls.user.save()

    def setUp(self):
        super(DRFTestCase, self).setUp()
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.client.logout()
        self.one = DummyModel.objects.create(title="One", code="one")
        self.two = DummyModel.objects.create(title="Two", code="two")

    def tearDown(self):
        self.client.logout()
        DummyModel.objects.all().delete()
        super(DRFTestCase, self).tearDown()

    def test_get_dummymodels(self):
        response = self.client.get("/api/dummies/")
        as_json_1 = response.json()

        # Drop to low level API so post_save does not trigger, meaning the
        # cached version is fetched on the next request.
        DummyModel.objects.filter(pk=self.one.pk).update(title="Onae")
        response = self.client.get("/api/dummies/")
        as_json_2 = response.json()
        self.assertEqual(as_json_1, as_json_2)

        # Modify it the normal way, which removes the item from cache.
        self.one.title = "Onbe"
        self.one.save()
        response = self.client.get("/api/dummies/")
        as_json_3 = response.json()
        self.assertNotEqual(as_json_1, as_json_3)

        # Trivial fetch to prove it is cached now
        response = self.client.get("/api/dummies/")
        as_json_4 = response.json()
        self.assertEqual(as_json_3, as_json_4)

        # Modify via API to confirm that post_save is fired implicitly
        data = {
            "title": "Onze"
        }
        response = self.client.patch("/api/dummies/%s/" % self.one.pk, data)
        response = self.client.get("/api/dummies/")
        as_json_5 = response.json()
        self.assertNotEqual(as_json_4, as_json_5)

        # Add an evaluate parameter which leads to a cache miss. Also
        # surreptiously edit the object so we can confirm a cache miss.
        DummyModel.objects.filter(pk=self.one.pk).update(title="Once")
        di = copy.deepcopy(settings.ULTRACACHE)
        di["drf"] = {"viewsets": {"*": {"evaluate": "request.user.is_anonymous"}}}
        with override_settings(ULTRACACHE=di):
            response = self.client.get("/api/dummies/")
            as_json_6 = response.json()
            self.assertNotEqual(as_json_5, as_json_6)

        # Trivial fetch to refresh the cache
        response = self.client.get("/api/dummies/")
        as_json_7 = response.json()

        # Disable viewset caching which leads to a cache miss. Also
        # surreptiously edit the object so we can confirm a cache miss.
        DummyModel.objects.filter(pk=self.one.pk).update(title="Onde")
        di = copy.deepcopy(settings.ULTRACACHE)
        di["drf"] = {"viewsets": {object: {}}}
        with override_settings(ULTRACACHE=di):
            response = self.client.get("/api/dummies/")
            as_json_8 = response.json()
            self.assertNotEqual(as_json_7, as_json_8)

        # Trivial fetch to refresh the cache
        response = self.client.get("/api/dummies/")
        as_json_9 = response.json()

        # Enable viewset caching for a single viewset class. Also
        # surreptiously edit the object so we can confirm a cache hit.
        DummyModel.objects.filter(pk=self.one.pk).update(title="Onee")
        di = copy.deepcopy(settings.ULTRACACHE)
        di["drf"] = {"viewsets": {DummyViewSet: {}}}
        with override_settings(ULTRACACHE=di):
            response = self.client.get("/api/dummies/")
            as_json_10 = response.json()
            self.assertEqual(as_json_9, as_json_10)

    def test_get_dummymodel(self):
        url = "/api/dummies/%s/" % self.one.pk
        response = self.client.get(url)
        as_json_1 = response.json()

        # Drop to low level API so post_save does not trigger, meaning the
        # cached version is fetched on the next request.
        DummyModel.objects.filter(pk=self.one.pk).update(title="Onfe")
        response = self.client.get(url)
        as_json_2 = response.json()
        self.assertEqual(as_json_1, as_json_2)

        # Modify it the normal way, which removes the item from cache.
        self.one.title = "Onge"
        self.one.save()
        response = self.client.get(url)
        as_json_3 = response.json()
        self.assertNotEqual(as_json_1, as_json_3)



import hashlib
from collections import OrderedDict

from django.core.cache import cache
from django.conf import settings
from django.http.cookie import SimpleCookie

from ultracache import _thread_locals


# The metadata itself can't be allowed to grow endlessly. This value is the
# maximum size in bytes of a metadata list. If your caching backend supports
# compression set a larger value.
try:
    MAX_SIZE = settings.ULTRACACHE["max-registry-value-size"]
except (AttributeError, KeyError):
    MAX_SIZE = 1000000

try:
    CONSIDER_HEADERS = [
        header.lower() for header in settings.ULTRACACHE["consider-headers"]
    ]
except (AttributeError, KeyError):
    CONSIDER_HEADERS = []

try:
    CONSIDER_COOKIES = [
        cookie.lower() for cookie in settings.ULTRACACHE["consider-cookies"]
    ]
except (AttributeError, KeyError):
    CONSIDER_COOKIES = []

# Raise on potentially confusing settings
if CONSIDER_COOKIES and ("cookie" in CONSIDER_HEADERS):
    raise RuntimeError(
        "consider-cookies has a value but cookie is also present in \
consider-headers"
    )


def reduce_list_size(li):
    """Return two lists
        - the last N items of li whose total size is less than MAX_SIZE
        - the rest of the original list li
    """
    # sys.getsizeof is nearly useless. All our data is stringable so rather
    # use that as a measure of size.
    size = len(repr(li))
    keep = li
    toss = []
    n = len(li)
    decrement_by = max(n / 10, 10)
    while (size >= MAX_SIZE) and (n > 0):
        n -= decrement_by
        toss = li[:-n]
        keep = li[-n:]
        size = len(repr(keep))
    return keep, toss


def cache_meta(recorder, cache_key, start_index=0, request=None):
    """Inspect request for objects in _ultracache and set appropriate entries
    in Django's cache."""

    path = None
    if request is not None:
        path = request.get_full_path()
        # todo: cache headers on the recorder since they never change during the
        # request.

        # Reduce headers to the subset as defined by the settings
        headers = OrderedDict()
        for k, v in sorted(request.META.items()):
            if (k == "HTTP_COOKIE") and CONSIDER_COOKIES:
                cookie = SimpleCookie()
                cookie.load(v)
                headers["cookie"] = "; ".join([
                    "%s=%s" % (k, morsel.value) for k, morsel \
                        in sorted(cookie.items()) if k in CONSIDER_COOKIES
                ])
            elif k.startswith("HTTP_"):
                k = k[5:].replace("_", "-").lower()
                if k in CONSIDER_HEADERS:
                    headers[k] = v

    # Lists needed for cache.get_many
    to_set_get_keys = []
    to_set_paths_get_keys = []
    to_set_content_types_get_keys = []
    to_set_content_types_paths_get_keys = []

    # Dictionaries needed for cache.set_many
    to_set = {}
    to_set_paths = {}
    to_set_content_types = {}
    to_set_content_types_paths = {}

    to_delete = []
    to_set_objects = []

    for ctid, obj_pk in recorder[start_index:]:
        # The object appears in these cache entries. If the object is modified
        # then these cache entries are deleted.
        key = "ucache-%s-%s" % (ctid, obj_pk)
        if key not in to_set_get_keys:
            to_set_get_keys.append(key)

        # The object appears in these paths. If the object is modified then any
        # caches that are read from when browsing to this path are cleared.
        key = "ucache-pth-%s-%s" % (ctid, obj_pk)
        if key not in to_set_paths_get_keys:
            to_set_paths_get_keys.append(key)

        # The content type appears in these cache entries. If an object of this
        # content type is created then these cache entries are cleared.
        key = "ucache-ct-%s" % ctid
        if key not in to_set_content_types_get_keys:
            to_set_content_types_get_keys.append(key)

        # The content type appears in these paths. If an object of this content
        # type is created then any caches that are read from when browsing to
        # this path are cleared.
        key = "ucache-ct-pth-%s" % ctid
        if key not in to_set_content_types_paths_get_keys:
            to_set_content_types_paths_get_keys.append(key)

        # A list of objects that contribute to a cache entry
        tu = (ctid, obj_pk)
        if tu not in to_set_objects:
            to_set_objects.append(tu)

    # todo: rewrite to handle absence of get_many
    di = cache.get_many(to_set_get_keys)
    for key in to_set_get_keys:
        v = di.get(key, None)
        keep = []
        if v is not None:
            keep, toss = reduce_list_size(v)
            if toss:
                to_set[key] = keep
                to_delete.extend(toss)
        if cache_key not in keep:
            if key not in to_set:
                to_set[key] = keep
            to_set[key] = to_set[key] + [cache_key]
    if to_set == di:
        to_set = {}

    di = cache.get_many(to_set_paths_get_keys)
    for key in to_set_paths_get_keys:
        v = di.get(key, None)
        keep = []
        if v is not None:
            keep, toss = reduce_list_size(v)
            if toss:
                to_set_paths[key] = keep
        if path is not None:
            if [path, headers] not in keep:
                if key not in to_set_paths:
                    to_set_paths[key] = keep
                to_set_paths[key] = to_set_paths[key] + [[path, headers]]
    if to_set_paths == di:
        to_set_paths = {}

    di = cache.get_many(to_set_content_types_get_keys)
    for key in to_set_content_types_get_keys:
        v = di.get(key, None)
        keep = []
        if v is not None:
            keep, toss = reduce_list_size(v)
            if toss:
                to_set_content_types[key] = keep
                to_delete.extend(toss)
        if cache_key not in keep:
            if key not in to_set_content_types:
                to_set_content_types[key] = keep
            to_set_content_types[key] = to_set_content_types[key] + [cache_key]
    if to_set_content_types == di:
        to_set_content_types = {}

    di = cache.get_many(to_set_content_types_paths_get_keys)
    for key in to_set_content_types_paths_get_keys:
        v = di.get(key, None)
        keep = []
        if v is not None:
            keep, toss = reduce_list_size(v)
            if toss:
                to_set_content_types_paths[key] = keep
        if path is not None:
            if [path, headers] not in keep:
               if key not in to_set_content_types_paths:
                    to_set_content_types_paths[key] = keep
               to_set_content_types_paths[key] = to_set_content_types_paths[key] \
                    + [[path, headers]]
    if to_set_content_types_paths == di:
        to_set_content_types_paths = {}

    # Deletion must happen first because set may set some of these keys
    if to_delete:
        try:
            cache.delete_many(to_delete)
        except NotImplementedError:
            for k in to_delete:
                cache.delete(k)

    # Do one set_many
    di = {}
    di.update(to_set)
    del to_set
    di.update(to_set_paths)
    del to_set_paths
    di.update(to_set_content_types)
    del to_set_content_types
    di.update(to_set_content_types_paths)
    del to_set_content_types_paths

    if to_set_objects:
        di[cache_key + "-objs"] = to_set_objects

    if di:
        try:
            cache.set_many(di, 86400)
        except NotImplementedError:
            for k, v in di.items():
                cache.set(k, v, 86400)


def get_current_site_pk(request):
    """Seemingly pointless function is so calling code doesn't have to worry
    about the import issues between Django 1.6 and later."""
    from django.contrib.sites.models import Site
    try:
        from django.contrib.sites.shortcuts import get_current_site
    except ImportError:
        from django.contrib.sites.models import get_current_site
    return get_current_site(request).pk


class EmptyMarker:
    pass
empty_marker_1 = EmptyMarker()
empty_marker_2 = EmptyMarker()


class Ultracache:
    """Cache arbitrary pieces of Python code.
    """

    def __init__(self, timeout, name, *params, request=None):
        self.timeout = timeout
        self.request = request
        self._cached = empty_marker_1
        s = ":".join([name] + [str(p) for p in params])
        hashed = hashlib.md5(s.encode("utf-8")).hexdigest()
        self.cache_key = "ucache-%s" % hashed
        if not hasattr(_thread_locals, "ultracache_recorder"):
            setattr(_thread_locals, "ultracache_recorder", [])
        self.start_index = len(_thread_locals.ultracache_recorder)
        self.used = False

    @property
    def cached(self):
        if self._cached is empty_marker_1:
            self._cached = cache.get(self.cache_key, empty_marker_2)
        return self._cached

    def __bool__(self):
        return self.cached is not empty_marker_2

    def cache(self, value):
        if self.used:
            raise RuntimeError(
                "The cache method may only be called once per Ultracache object."
            )
        cache.set(self.cache_key, value, self.timeout)
        cache_meta(
            _thread_locals.ultracache_recorder,
            self.cache_key,
            start_index=self.start_index,
            request=self.request
        )
        self.used = True
